"""piwrapper - Python connector access OSISOFT PI server via PI Web API."""
from piwrapper.PiConnect import PIValue, Connection
from piwrapper.PiConsts import BufferOption, UpdateOption, SummaryType


__version__ = '0.1.0'
__author__ = 'Josh Chen <jcpythonlib@gmail.com>'
__all__ = []



from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

class BufferOption(Enum):
    """
    Enumeration for buffer option type documented here:
    https://docs.osisoft.com/bundle/pi-web-api-reference/page/help/topics/buffer-option.html
    """

    DO_NO_BUFFER = "DoNotBuffer"
    BUFFER_IF_POSSIBLE = "BufferIdPossible"
    BUFFER = "Buffer"

class UpdateOption(Enum):
    """
    Enumeration for Update Option as documented here:
    https://docs.osisoft.com/bundle/pi-web-api-reference/page/help/topics/update-option.html
    """

    REPLACE = "Replace"
    INSERT = "Insert"
    NO_REPLACE = "NoReplace"
    REPLACE_ONLY = "ReplaceOnly"
    INSERT_NO_COMPRESSION = "InsertNoCompression"
    REMOVE = "Remove"

class SummaryType(Enum):
    """
    Enumeration for summary type documented here:
    https://docs.osisoft.com/bundle/pi-web-api-reference/page/help/topics/summary-type.html
    """

    TOTAL = "Total"
    AVERAGE = "Average"
    MINIMUM = "Minimum"
    MAXIMUM = "Maximum"
    RANGE = "Range"
    STD_DEV = "StdDev"
    POPULATION_STD_DEV = "PopulationStdDev"
    COUNT = "Count"
    PERCENT_GOOD = "PercentGood"
    ALL = "All"
    ALL_FOR_NON_NUMERIC = "AllForNonNumeric"





import datetime
import json
import requests
import urllib3
import pandas as pd
from dataclasses import dataclass
from requests_kerberos import HTTPKerberosAuth, OPTIONAL, REQUIRED
from typing import Any, Dict, List, Optional, Tuple
from piwrapper.PiConsts import BufferOption, UpdateOption


@dataclass
class PIValue:
    """
    Represents a single PI value
    #two ways to pass timestamp
    1 datetime.datetime(2021,11,13,21,00)
    2 import dateutil
    dateutil.parser.parse("2021-11-13T21:00:00")
    """

    time_stamp: datetime.datetime = datetime.datetime.min
    units_abbreviation: Optional[str] = None
    good: Optional[bool] = None
    questionable: Optional[bool] = None
    value: Optional[Any] = None

    def to_json(self) -> str:
        """
        Serialize to a json string
        :return: json representation of the object
        """
        tmp_dict = {"Timestamp": self.time_stamp.isoformat()}
        if self.units_abbreviation is not None:
            tmp_dict["UnitsAbbreviation"] = self.units_abbreviation
        if self.good is not None:
            tmp_dict["Good"] = self.good
        if self.questionable is not None:
            tmp_dict["Questionable"] = self.questionable
        if self.value is not None:
            tmp_dict["Value"] = self.value
        return json.dumps(obj=tmp_dict)

class Connection:
    """
    Pi class for connecting to the PI REST API
    """

    def __init__(
        self,
        server: str,
        basic_credentials: Optional[Tuple[str, str]] = None,
        verify: bool = True,
    ):
        """
        Initialises connection instance

        :param basic_credentials: Optional tuple of username password strings if basic authentication is used
        :param server: URI for the PI REST API server
        :verify: Verify the certificate of the server before connecting
        """

        if basic_credentials is None:
            self.auth = HTTPKerberosAuth(
                mutual_authentication=REQUIRED if verify else OPTIONAL
            )
        else:
            self.auth = basic_credentials

        self.base_url: str = server
        self.starting_url: str = f"""https://{self.base_url}/piwebapi/"""
        self.verify: bool = verify


        if self.verify is False:
            # Disable the insecure request warnings if the user doesn't want to check URL certificates
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

    def get_data_servers(self) -> list:
        """
        Get all dataservers
        :return: list of data servers
        raise: connectionError: fail to connect to Pi server
        """
        response: requests.Response = requests.get(
            self.starting_url,
            auth=self.auth,
            headers={"Content-Type": "application/json"},
            verify=self.verify,
        )

        if response.status_code != requests.codes.ok:
            raise ConnectionError("Connection to PI REST API Server failed")

        top_level_content = json.loads(s=response.content)

        # Get data server
        data_servers_url = top_level_content["Links"]["DataServers"]
        data_server_response: requests.Response = requests.get(
            data_servers_url,
            auth=self.auth,
            headers={"Content-Type": "application/json"},
            verify=self.verify,
        )
        data_server_content_list = json.loads(s=data_server_response.content)
        self.data_server_list: List[Dict[str, Any]] = [
            data_server_content
            for data_server_content in data_server_content_list["Items"]
        ]
        return self.data_server_list
    def _find_pi_webid(
        self,
        pi_tag: str) ->dict:
        """
        Search associated webid for pi_tag
        :param: pi_tag: full or partial pi tag name
        :return: dictionary with tag name as key and webid as value
        :raise LookupError: No matching tag can be found
        :raise ValueError: No valid response returned
        """
        pi_tag_url: str = f"""{self.starting_url}search/query?q=name:{pi_tag}"""
        response: requests.Response = requests.get(
            pi_tag_url,
            auth=self.auth,
            headers={"Content-Type": "application/json"},
            verify=self.verify,
        )
        if response.status_code != requests.codes.ok:
            raise LookupError(response)
        response_dict : Dict[str, Any] =json.loads(s=response.content)
        if response_dict["Items"]:
            webid_dic = {idx["Name"]: idx["WebID"] for idx in response_dict["Items"] }
        else:
            raise ValueError("No response. Please check tag name")

        return webid_dic
    def _single_interpolated_value_getter(
        self,
        webid: str):
        """
        get interpolated value in data frame for single webid
        :param webid: webid
        :return: interpolated value in a dataframe
        :raise LookupError: No matching tag can be found
        :raise ValueError: No response with the webid.
       """        
        resource_url: str = f"""{self.starting_url}streams/{webid}/interpolated"""
        response: requests.Response = requests.get(
            resource_url,
            auth=self.auth,
            headers={"Content-Type": "application/json"},
            verify=self.verify,
        )

        if response.status_code != requests.codes.ok:
            raise LookupError(response)
        response_dict : Dict[str, Any] =json.loads(s=response.content)
        if response_dict["Items"]:
            response_df = pd.DataFrame(response_dict["Items"])
        else:
            raise ValueError("No response. Please check webid")
        return response_df

    def get_interpolated_value(   
        self,
        pi_tag: str):
        """
        get pi point interpolated value
        :param pi_tag: name of pi tag
        :return: interpolated value in a dataframe
        :raise LookupError: No matching tag can be found
        :raise ValueError: more than one tags foud
        """
        webid_dic: Dict[str, Any]  = self._find_pi_webid(pi_tag)
        if len(webid_dic) > 1:
            raise ValueError(f"more than one tags have been found: {webid_dic.keys()}, use get_interpolated_values method instead")
        _, v = next(iter(webid_dic.items()))
        response_df=self._single_interpolated_value_getter(v)
        return response_df

    def get_interpolated_values(   
        self,
        pi_tag: str) -> dict:
        """
        get pi point interpolated value
        :param pi_tag: name of pi tag
        :return: nested interpolated value in the DICTIONARY 
        with tag name as key as interpolated value as value
        :raise LookupError: No matching tag can be found
        """
        webid_dic: Dict[str, Any]  = self._find_pi_webid(pi_tag)
        tag_names = list(webid_dic.keys())
        response_dict : Dict[str, Any] = {tag: self._single_interpolated_value_getter(webid_dic[tag]) for tag in tag_names }

        return response_dict
    def update_value(
        self,
        value: PIValue,
        update_option: UpdateOption,
        buffer_option: BufferOption,
        pi_tag: str = None,
        webid: str = None
    ) -> str:
        """
        Update a value for the specified pi tag

        :param value: value to update with Dataclass attributes
        :param pi_tag: update the value for this pi_tag
        :param update_option: update mode to apply
        :param buffer_option: buffering more to apply
        :param web_id: use web_id to update 
        :return: Location or the added object if successful
        :raises ValueError: cannot have both webid and pi_tag at the same time
        :raises ValueError: cannot have more than one webid
        :raises LookupError: Failed to post an update to the current web_id
        """

        if webid is not None and pi_tag is not None:
            raise ValueError("Cannot pass both webid and pi_tag at the same time")
        if webid == None:
            webid_dic: Dict[str, Any]  = self._find_pi_webid(pi_tag)
            if len(webid_dic) > 1:
                raise ValueError(f"more than one associated webids have been found: {webid_dic.keys()}, use single webid instead")
            _, v = next(iter(webid_dic.items()))
            url: str = (f"{self.starting_url}streams/{v}/value?updateOption={update_option.value}&bufferOption={buffer_option.value}")
        else:
            url: str = (f"{self.starting_url}streams/{webid}/value?updateOption={update_option.value}&bufferOption={buffer_option.value}")        
        response: requests.Response = requests.post(
            url=url,
            data=value.to_json(),
            auth=self.auth,
            headers={"Content-Type": "application/json"},
            verify=self.verify,
        )


        if response.status_code != requests.codes.ok and response.status_code != requests.codes.no_content:
            raise LookupError(response)

        return response.headers["Location"]








import io
import os
import re

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())


setup(
    name="piwrapper",
    version="0.1.0",
    url="https://github.com/joshylchen/piwrapper",
    license='MIT',

    author="Josh Chen",
    author_email="jcpythonlib@gmail.com",

    description="Python connector access OSISOFT PI server via PI Web API.",
    long_description=read("README.rst"),

    packages=find_packages(exclude=('tests',)),

    install_requires=[
        "numpy>=1.21.3",
        "pandas>=1.3.4",
        "python-dateutil>=2.8",
        "requests>=2.26",
        "requests-kerberos>=0.12",
        "tqdm>=4.32",
        "tzlocal>=1.5",
        "urllib3>=1.26.7",],
    python_requires=">=3.6.0",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',        
    ],
)
